import os
from datetime import timedelta
from pathlib import Path
from typing import Callable

from airflow import DAG
from kubernetes import client

from dataverk_airflow.kubernetes_operator import kubernetes_operator, VALID_PYTHON_VERSIONS


def notebook_operator(
        dag: DAG,
        name: str,
        nb_path: str,
        repo: str = None,
        log_output: bool = False,
        image: str = None,
        branch: str = "main",
        email: str = None,
        slack_channel: str = None,
        extra_envs: dict = {},
        allowlist: list = [],
        requirements_path: str = None,
        python_version: str = "3.11",
        resources: client.V1ResourceRequirements = None,
        startup_timeout_seconds: int = None,
        retries: int = None,
        delete_on_finish: bool = True,
        retry_delay: timedelta = None,
        do_xcom_push: bool = False,
        on_success_callback: Callable = None,
        container_uid: int = 50000,
        use_uv_pip_install: bool = False,
        override_notebook_kernelspec: bool = True,
        **kwargs,
):
    """Operator for executing Jupyter notebooks.

    :param dag: DAG: owner DAG
    :param name: str: Name of task
    :param repo: str: Github repo
    :param nb_path: str: Path to notebook in repo
    :param log_output: bool: Write logs from notebook to stdout, default False
    :param image: str: Dockerimage the pod should use
    :param branch: str: Branch in repo, default "main"
    :param email: str: Email of owner
    :param slack_channel: str: Name of Slack channel, default None (no Slack notification)
    :param extra_envs: dict: dict with environment variables example: {"key": "value", "key2": "value2"}
    :param allowlist: list: list of hosts and port the task needs to reach on the format host:port
    :param requirements_path: bool: Path (including filename) to your requirements.txt
    :param python_version: str: desired Python version for the environment your code will be running in when using the default image. We offer only supported versions of Python, and default to the latest version if this parameter is omitted. See https://devguide.python.org/versions/ for available versions.
    :param resources: dict: Specify cpu and memory resource usage (dict: request/limit: {"memory": "", "cpu": "", "ephemeral-storage": ""}), default None
    :param startup_timeout_seconds: int: pod startup timeout
    :param retries: int: Number of retries for task before DAG fails, default 3
    :param delete_on_finish: bool: Whether to delete pod on completion
    :param retry_delay: timedelta: Time inbetween retries, default 5 seconds
    :param do_xcom_push: bool: Enable xcom push of content in file '/airflow/xcom/return.json', default False
    :param container_uid: int: User ID for the container image. Root (id = 0) is not allowed, defaults to 50000 (standard uid for airflow).
    :param on_success_callback: Callable
    :param use_uv_pip_install: bool: Use uv pip install, default False
    :param override_notebook_kernelspec: bool: Whether to override the kernelspec in the notebook metadata, default True (will ensure that python3 kernel is used)

    :return: KubernetesPodOperator
    """
    if not image:
        if python_version not in VALID_PYTHON_VERSIONS:
            raise ValueError(
                f"When using the default image the python_version argument must be set to "
                f"{', '.join(VALID_PYTHON_VERSIONS[:-1])} or {VALID_PYTHON_VERSIONS[-1]}: value '{python_version}' is not supported" 
            )
        try:
            image = os.environ["DATAVERK_IMAGE_PYTHON_" + python_version.replace(".","")]
        except KeyError:
            image = os.getenv("KNADA_AIRFLOW_OPERATOR_IMAGE")

    # Needs to override python3 kernelspec if using uv
    if use_uv_pip_install:
        cmds = [f"python -m ipykernel install --user --name=python3 --display-name 'Python (UV)' && papermill {Path(nb_path).name} output.ipynb"]
    else:
        cmds = [f"papermill {Path(nb_path).name} output.ipynb"]

    if log_output:
        cmds[-1] += " --log-output"
    if override_notebook_kernelspec:
        cmds[-1] += " --kernel python3"

    kwargs = {
        "dag": dag, "name": name, "repo": repo, "entrypoint": None, "image": image, "cmds": cmds, "branch": branch, "email": email,
        "slack_channel": slack_channel, "extra_envs": extra_envs, "allowlist": allowlist, "requirements_path": requirements_path,
        "resources": resources, "startup_timeout_seconds": startup_timeout_seconds, 
        "retries": retries, "delete_on_finish": delete_on_finish, "retry_delay": retry_delay, "do_xcom_push": do_xcom_push,
        "on_success_callback": on_success_callback, "working_dir": str(Path(nb_path).parent), "container_uid": container_uid, "use_uv_pip_install": use_uv_pip_install,
        **kwargs,
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    return kubernetes_operator(**kwargs)



import os
from datetime import datetime
from typing import List, Union

from airflow.operators.email import EmailOperator
from airflow.providers.slack.operators.slack import SlackAPIPostOperator


def create_email_notification(dag, email: Union[List[str], str], name: str, namespace: str):
    return EmailOperator(
                dag=dag,
                task_id="airflow_task_failed_email",
                to=email,
                subject=f"Airflow task {name} error",
                html_content=f"<p> Airflow task {name} i DAG {dag._dag_id} feiler i namespace {namespace} "
                f"kl. {datetime.now().isoformat()}. ",
            )


def create_slack_notification(dag, slack_channel: str, name: str, namespace: str):
    return SlackAPIPostOperator(
            dag=dag,
            task_id="airflow_task_failed_slack",
            slack_conn_id="slack_connection",
            text=f":error: DAG *{dag._dag_id}* feilet :error: \n Sjekk task _{name}_ i namespace {namespace}",
            channel=slack_channel,
    )



import os
from datetime import timedelta
from typing import Callable, List, Tuple

from airflow import DAG

from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator
from airflow.providers.cncf.kubernetes.utils.pod_manager import OnFinishAction
from kubernetes import client
from kubernetes.client.models import (
    V1ConfigMapVolumeSource,
    V1Container,
    V1PodSecurityContext,
    V1SeccompProfile,
    V1SecretVolumeSource,
    V1Volume,
    V1VolumeMount,
)

from dataverk_airflow import git_clone, bucket_read
from dataverk_airflow.notifications import (
    create_email_notification,
    create_slack_notification,
)

CA_BUNDLE_PATH = "/etc/pki/tls/certs/ca-bundle.crt"
POD_WORKSPACE_DIR = "/workspace"

VALID_PYTHON_VERSIONS = ["3.9", "3.10", "3.11", "3.12", "3.13"]


class MissingValueException(Exception):
    """Raised when a required value is missing."""


def kubernetes_operator(
        dag: DAG,
        name: str,
        image: str,
        repo: str = None,
        entrypoint: List[str] = None,
        cmds: list = [],
        branch: str = "main",
        email: str = None,
        slack_channel: str = None,
        extra_envs: dict = {},
        allowlist: list = [],
        requirements_path: str = None,
        resources: client.V1ResourceRequirements = None,
        startup_timeout_seconds: int = 360,
        retries: int = 3,
        delete_on_finish: bool = True,
        retry_delay: timedelta = timedelta(seconds=5),
        do_xcom_push: bool = False,
        container_uid: int = 50000,
        on_success_callback: Callable = None,
        working_dir: str = None,
        use_uv_pip_install: bool = False,
        **kwargs,
):
    """Simplified operator for creating KubernetesPodOperator.

    If you use the slack_channel argument, the following host will also be added:
    - hooks.slack.com

    If you use the email argument, the following host will also be added:
    - smtp-address

    If you use the requirements_path argument, the following hosts will also be added:
    - pypi.org
    - files.pythonhosted.org
    - pypi.python.org

    :param dag: DAG: owner DAG
    :param name: str: Name of task
    :param repo: str: Github repo
    :param image: str: Dockerimage the pod should use
    :param cmds: str: Command to run in pod
    :param branch: str: Branch in repo, default "main"
    :param email: str: Email of owner
    :param slack_channel: str: Name of Slack channel, default None (no Slack notification)
    :param extra_envs: dict: dict with environment variables example: {"key": "value", "key2": "value2"}
    :param allowlist: list: list of hosts and port the task needs to reach on the format host:port
    :param requirements_path: bool: Path (including filename) to your requirements.txt
    :param resources: dict: Specify cpu and memory resource usage (dict: request/limit: {"memory": "", "cpu": "", "ephemeral-storage": ""}), default ephemeral-storage: 100Mi
    :param startup_timeout_seconds: int: pod startup timeout
    :param retries: int: Number of retries for task before DAG fails, default 3
    :param delete_on_finish: bool: Whether to delete pod on completion
    :param retry_delay: timedelta: Time inbetween retries, default 5 seconds
    :param do_xcom_push: bool: Enable xcom push of content in file '/airflow/xcom/return.json', default False
    :param container_uid: int: User ID for the container image. Root (id = 0) is not allowed, defaults to 50000 (standard uid for airflow).
    :param on_success_callback: a function or list of functions to be called when a task instance
        of this task fails. a context dictionary is passed as a single
        parameter to this function. Context contains references to related
        objects to the task instance and is documented under the macros
        section of the API.
    :param working_dir: str: Path to working directory
    :param use_uv_pip_install: bool: Use uv pip install, default False

    :return: KubernetesPodOperator
    """
    is_composer = True if os.getenv("GCS_BUCKET") else False

    if not is_composer:
        if repo in [None, ""]:
            raise MissingValueException("repo cannot be empty")
        if container_uid == 0:
            raise ValueError(f"container_uid is not allowed to be 0 (root user)")

    if image == "":
        raise MissingValueException("image cannot be empty")

    if slack_channel:
        allowlist.append("hooks.slack.com")

    if email:
        allowlist.append(os.getenv("AIRFLOW__SMTP__SMTP_HOST") +
                         ":"+os.getenv("AIRFLOW__SMTP__SMTP_PORT"))

    namespace = get_namespace(is_composer)

    def on_failure(context):
        if email:
            send_email = create_email_notification(dag, email, name, namespace)
            send_email.execute(context)

        if slack_channel:
            slack_notification = create_slack_notification(
                dag, slack_channel, name, namespace)
            slack_notification.execute(context)

    on_finish_action = OnFinishAction.DELETE_POD if delete_on_finish else OnFinishAction.KEEP_POD

    if working_dir:
        working_dir = POD_WORKSPACE_DIR + "/" + working_dir
    else:
        working_dir = POD_WORKSPACE_DIR

    container_command = entrypoint if entrypoint else ["/bin/sh", "-c"]

    if requirements_path:
        if use_uv_pip_install:
            cmds = [
                f"uv venv .local && . .local/bin/activate && uv pip install -r {POD_WORKSPACE_DIR}/{requirements_path} --no-cache-dir"] + cmds
        else:
            cmds = [
                f"pip install -r {POD_WORKSPACE_DIR}/{requirements_path} --user --no-cache-dir"] + cmds

        allowlist.append("pypi.org")
        allowlist.append("files.pythonhosted.org")
        allowlist.append("pypi.python.org")

    if resources is None:
        resources = client.V1ResourceRequirements(
            requests={
                "ephemeral-storage": "100Mi",
            }
        )

    return KubernetesPodOperator(
        dag=dag,
        on_failure_callback=on_failure,
        name=name,
        task_id=name,
        startup_timeout_seconds=startup_timeout_seconds,
        on_finish_action=on_finish_action,
        annotations={
          "allowlist": ",".join(allowlist),
          "cluster-autoscaler.kubernetes.io/safe-to-evict": "false",
        },
        image=image,
        env_vars=env_vars(is_composer, extra_envs),
        config_file=config_file(is_composer),
        do_xcom_push=do_xcom_push,
        container_resources=resources,
        retries=retries,
        on_success_callback=on_success_callback,
        retry_delay=retry_delay,
        executor_config={
            "pod_override": client.V1Pod(
                metadata=client.V1ObjectMeta(annotations={"allowlist": ",".join(allowlist)}),
            ),
        },
        full_pod_spec=client.V1Pod(
            spec=client.V1PodSpec(
                containers=[
                    client.V1Container(
                        name="base",
                        working_dir=working_dir,
                        security_context=client.V1SecurityContext(
                            allow_privilege_escalation=False,
                            run_as_user=container_uid,
                        ) if not is_composer else None
                    )
                ]
            )
        ),
        init_containers=init_containers(is_composer, repo, branch, container_uid),
        image_pull_secrets=os.getenv("K8S_IMAGE_PULL_SECRETS"),
        labels={
            "component": "worker",
            "release": "airflow"
        },
        cmds=container_command,
        arguments=[" && ".join(cmds)] if cmds is not None and len(cmds) > 0 else None,
        volume_mounts=volume_mounts(is_composer),
        service_account_name=os.getenv("TEAM", "default"),
        volumes=volumes(is_composer),
        security_context=V1PodSecurityContext(
            fs_group=0,
            run_as_non_root=True,
            seccomp_profile=V1SeccompProfile(
                type="RuntimeDefault"
            )
        ) if not is_composer else None,
        **kwargs,
    )


def get_namespace(is_composer: bool) -> str:
    if is_composer:
        return "composer-user-workloads"
    else:
        return os.getenv("NAMESPACE")


def env_vars(is_composer: bool, extra_envs: dict) -> dict:
    env_vars = {
        "NLS_LANG": "NORWEGIAN_NORWAY.AL32UTF8",
        "TZ": os.getenv("TZ", "Europe/Oslo"),
        **extra_envs
    }

    if not is_composer:
        if not os.getenv("INTEGRATION_TEST"):
            env_vars["REQUESTS_CA_BUNDLE"] = CA_BUNDLE_PATH

        env_vars["KNADA_TEAM_SECRET"] = os.environ["KNADA_TEAM_SECRET"]

    return env_vars


def config_file(is_composer: bool) -> str:
    if not os.getenv("INTEGRATION_TEST"):
        return "/home/airflow/composer_kube_config" if is_composer else None


def init_containers(is_composer: bool, repo: str, branch: str, run_as_user: str) -> List[V1Container]:
    if is_composer:
        return [
            bucket_read(POD_WORKSPACE_DIR)
        ]
    else:
        return [
            git_clone(repo, branch, POD_WORKSPACE_DIR, run_as_user)
        ]


def volume_mounts(is_composer: bool) -> List[V1VolumeMount]:
    volume_mounts = [
        V1VolumeMount(
            name="dags-data",
            mount_path=POD_WORKSPACE_DIR,
            sub_path=None,
            read_only=False
        )
    ]

    if not is_composer:
        volume_mounts.append(
            V1VolumeMount(
                name="ca-bundle-pem",
                mount_path=CA_BUNDLE_PATH,
                read_only=True,
                sub_path="ca-bundle.pem"
            )
        )

    return volume_mounts


def volumes(is_composer: bool) -> List[V1Volume]:
    volumes = [
        V1Volume(
            name="dags-data"
        )
    ]

    if not is_composer:
        volumes.append(
            V1Volume(
                name="airflow-git-secret",
                secret=V1SecretVolumeSource(
                    default_mode=448,
                    secret_name=os.getenv(
                        "K8S_GIT_CLONE_SECRET", "github-app-secret"),
                )
            )
        )
        volumes.append(
            V1Volume(
                name="ca-bundle-pem",
                config_map=V1ConfigMapVolumeSource(
                    default_mode=420,
                    name="ca-bundle-pem",
                )
            )
        )

    return volumes



import os
import kubernetes.client as k8s

def bucket_read(
    mount_path: str,
) -> k8s.V1Container:
    return k8s.V1Container(
        name="read-bucket",
        image="google/cloud-sdk:alpine",
        volume_mounts=[
            k8s.V1VolumeMount(
                name="dags-data",
                mount_path=mount_path,
                sub_path=None,
                read_only=False
            ),
        ],
        command=["/bin/sh", "-c"],
        args=[
            f"gsutil rsync -r -x '^logs/*|^.*-build/*' gs://{os.environ['GCS_BUCKET']} {mount_path} && chmod -R 777 {mount_path}"
        ],
        resources=k8s.V1ResourceRequirements(
            requests={
                "memory": "2Gi", 
                "cpu": "500m",
                "ephemeral-storage": "2Gi",
            },
        ),
    )
